#!/usr/bin/env python3
# -*- coding:utf-8 -*-
################################################################
# Copyright 2025 Dong Zhaorui. All rights reserved.
# Author: Dong Zhaorui 847235539@qq.com
# Date  : 2025-11-04
################################################################

import numpy as np


class HexCtrlUtilMitJoint:

    def __init__(self, ctrl_limit: np.ndarray | None = None):
        if ctrl_limit is not None:
            self.__ctrl_upper = ctrl_limit.copy()
            self.__ctrl_lower = -ctrl_limit.copy()
        else:
            self.__ctrl_upper = None
            self.__ctrl_lower = None

    def __call__(self, kp, kd, q_tar, dq_tar, q_cur, dq_cur, tau_comp):
        q_err = q_tar - q_cur
        dq_err = dq_tar - dq_cur
        tau_ctrl = kp * q_err + kd * dq_err
        if (self.__ctrl_upper is not None) and (self.__ctrl_lower is not None):
            tau_ctrl = np.clip(tau_ctrl, self.__ctrl_lower, self.__ctrl_upper)
        return tau_ctrl + tau_comp


class HexCtrlUtilPidJoint:

    def __init__(
        self,
        kp: np.ndarray,
        ki: np.ndarray,
        kd: np.ndarray,
        ctrl_limit: np.ndarray | None = None,
    ):
        self.__kp = kp.copy()
        self.__ki = ki.copy()
        self.__kd = kd.copy()
        if ctrl_limit is not None:
            self.__ctrl_upper = ctrl_limit.copy()
            self.__ctrl_lower = -ctrl_limit.copy()
        else:
            self.__ctrl_upper = None
            self.__ctrl_lower = None

        # buffer
        self.__last_ctrl = np.zeros_like(kp)
        self.__last_q_err = np.zeros_like(kp)
        self.__last_dq_err = np.zeros_like(kp)

    def __call__(self, q_tar, q_cur, dq_tar, dq_cur, tau_comp):
        q_err = q_tar - q_cur
        dq_err = dq_tar - dq_cur

        p_term = self.__kp * (q_err - self.__last_q_err)
        i_term = self.__ki * q_err
        d_term = self.__kd * (dq_err - self.__last_dq_err)
        delta_ctrl = p_term + i_term + d_term
        tau_ctrl = self.__last_ctrl + delta_ctrl
        if (self.__ctrl_upper is not None) and (self.__ctrl_lower is not None):
            tau_ctrl = np.clip(tau_ctrl, self.__ctrl_lower, self.__ctrl_upper)

        self.__last_ctrl = tau_ctrl
        self.__last_q_err = q_err
        self.__last_dq_err = dq_err
        return tau_ctrl + tau_comp


class HexCtrlUtilIntJoint:

    def __init__(
        self,
        ki,
        dt,
        limit=np.array([1.0, 1.0, 1.0, 1.0, 1.0, 1.0]),
        near_ratio=0.5,
        near_threshold=1e-1,
    ):
        # constants
        self.__ki = ki.copy()
        self.__dt = dt
        self.__limit_upper = np.array(
            [limit.copy(), limit.copy() * near_ratio])
        self.__limit_lower = np.array(
            [-limit.copy(), -limit.copy() * near_ratio])
        self.__near_threshold = near_threshold

        # variables
        self.__i_term = np.zeros(limit.shape[0])

    def __call__(self, cur_q, tar_q):
        err = tar_q - cur_q
        self.__i_term += self.__ki * err * self.__dt

        # limit idx
        col = np.arange(err.shape[0])
        row = (np.fabs(err) < self.__near_threshold).astype(int)

        # i_term limit
        self.__i_term = np.clip(
            self.__i_term,
            self.__limit_lower[row, col],
            self.__limit_upper[row, col],
        )
        return self.__i_term


class HexCtrlUtilMitWork:

    def __init__(self, ctrl_limit: np.ndarray | None = None):
        if ctrl_limit is not None:
            self.__ctrl_upper = ctrl_limit.copy()
            self.__ctrl_lower = -ctrl_limit.copy()
        else:
            self.__ctrl_upper = None
            self.__ctrl_lower = None

    def __call__(self, kp, kd, se3_tar, dse3_tar, se3_cur, dse3_cur, tau_comp):
        se3_err = se3_tar - se3_cur
        dse3_err = dse3_tar - dse3_cur
        tau_ctrl = kp * se3_err + kd * dse3_err
        if (self.__ctrl_upper is not None) and (self.__ctrl_lower is not None):
            tau_ctrl = np.clip(tau_ctrl, self.__ctrl_lower, self.__ctrl_upper)
        return tau_ctrl + tau_comp


class HexCtrlUtilIntWork:

    def __init__(
        self,
        ki,
        dt,
        limit=np.array([1.0, 1.0, 1.0, 1.0, 1.0, 1.0]),
        near_ratio=0.5,
        near_threshold=1e-1,
    ):
        # constants
        self.__ki = ki.copy()
        self.__dt = dt
        self.__limit_upper = np.array(
            [limit.copy(), limit.copy() * near_ratio])
        self.__limit_lower = np.array(
            [-limit.copy(), -limit.copy() * near_ratio])
        self.__near_threshold = near_threshold

        # variables
        self.__i_term = np.zeros(limit.shape[0])

    def __call__(self, cur_se3, tar_se3):
        se3_err = tar_se3 - cur_se3
        self.__i_term += self.__ki * se3_err * self.__dt

        # limit idx
        col = np.arange(se3_err.shape[0])
        row = (np.fabs(se3_err) < self.__near_threshold).astype(int)

        # i_term limit
        self.__i_term = np.clip(
            self.__i_term,
            self.__limit_lower[row, col],
            self.__limit_upper[row, col],
        )
        return self.__i_term



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
################################################################
# Copyright 2025 Dong Zhaorui. All rights reserved.
# Author: Dong Zhaorui 847235539@qq.com
# Date  : 2025-10-11
################################################################

import time
import threading
import queue
import h5py, hdf5plugin
import numpy as np


class HexHdf5Writer:

    def __init__(self, file_path: str):
        hdf5plugin.register()
        self.__file_path = file_path
        self.__hdf5_file = h5py.File(file_path, "w", libver='latest')
        self.__group_dict = {}
        self.__dataset_dict = {}

        self.__queue = queue.Queue(maxsize=1024)
        self.__stop_event = threading.Event()
        self.__batch_size = 32
        self.__writer_cnt = 0
        self.__writer_thread = None
        self.__writer_exc = None

    def __del__(self):
        self.stop()

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()

    def start(self):
        if self.__writer_thread and self.__writer_thread.is_alive():
            return
        self.__stop_event.clear()
        self.__writer_thread = threading.Thread(
            target=self.__writer_loop,
            daemon=True,
        )
        self.__writer_thread.start()

    def stop(self):
        self.__stop_event.set()
        self.__writer_thread.join()
        self.summary()
        if self.__hdf5_file is not None:
            try:
                self.__hdf5_file.flush()
                self.__hdf5_file.close()
            except Exception:
                pass
            self.__hdf5_file = None
        if self.__writer_exc:
            raise self.__writer_exc

    def get_shape(self, group_name: str, dataset_name: str = "data"):
        dataset = self.__get_dataset_handle(group_name, dataset_name)
        return dataset.shape

    def get_dtype(self, group_name: str, dataset_name: str = "data"):
        dataset = self.__get_dataset_handle(group_name, dataset_name)
        return dataset.dtype

    def __get_dataset_handle(
        self,
        group_name: str,
        dataset_name: str = "data",
    ):
        dataset_key = f"{group_name}/{dataset_name}"
        if dataset_key not in self.__dataset_dict:
            if group_name not in self.__hdf5_file:
                raise KeyError(f"Group '{group_name}' not found in HDF5 file")
            if dataset_name not in self.__hdf5_file[group_name]:
                raise KeyError(
                    f"Dataset '{dataset_name}' not found in group '{group_name}'"
                )
            self.__dataset_dict[dataset_key] = self.__hdf5_file[group_name][
                dataset_name]
        return self.__dataset_dict[dataset_key]

    def summary(self):
        if self.__hdf5_file is None:
            print("HDF5 file is closed. Cannot get summary.")
            return

        print("#" * 50)
        print(f"HDF5 File: {self.__file_path}")
        print("#" * 50)

        for group_name in self.__hdf5_file.keys():
            print("-" * 30)
            print(f"Group: {group_name}")
            print("-" * 30)
            print(
                f"get_delta: {(self.__hdf5_file[group_name]['get_ts'][-1] - self.__hdf5_file[group_name]['get_ts'][0]) / 1e9}s"
            )
            print(
                f"sen_delta: {(self.__hdf5_file[group_name]['sen_ts'][-1] - self.__hdf5_file[group_name]['sen_ts'][0]) / 1e9}s"
            )

            dtype = self.get_dtype(group_name)
            shape = self.get_shape(group_name)
            print(f"  Dataset: {group_name}")
            print(f"    Dtype: {dtype}")
            print(f"    Shape: {shape}")

        print("#" * 50)

    def __writer_loop(self):
        try:
            batch = []
            while not (self.__stop_event.is_set() and self.__queue.empty()):
                try:
                    item = self.__queue.get(timeout=0.1)
                    batch.append(item)
                    if len(batch) >= self.__batch_size:
                        self.__flush_batch(batch)
                        batch.clear()
                    continue
                except queue.Empty:
                    if batch:
                        self.__flush_batch(batch)
                        batch.clear()
                    continue
            if batch:
                self.__flush_batch(batch)
        except Exception as e:
            self.__writer_exc = e
            raise

    def __flush_batch(self, batch):
        buckets = {}
        for group, data, gts, sts in batch:
            buckets.setdefault(group, []).append((data, gts, sts))

        for group, items in buckets.items():
            dataset_key = f"{group}/data"
            get_ts_key = f"{group}/get_ts"
            sen_ts_key = f"{group}/sen_ts"
            ds = self.__dataset_dict[dataset_key]
            d_get = self.__dataset_dict[get_ts_key]
            d_sen = self.__dataset_dict[sen_ts_key]

            # 将 items 合并成 numpy arrays
            data_arr = np.stack([it[0] for it in items], axis=0)
            gts_arr = np.stack([it[1] for it in items], axis=0).reshape(-1, 1)
            sts_arr = np.stack([it[2] for it in items], axis=0).reshape(-1, 1)

            n_old = ds.shape[0]
            n_new = n_old + data_arr.shape[0]
            # 一次 resize（比逐帧 resize 快得多）
            ds.resize((n_new, *ds.shape[1:]))
            d_get.resize((n_new, 1))
            d_sen.resize((n_new, 1))

            ds[n_old:n_new, ...] = data_arr
            d_get[n_old:n_new, :] = gts_arr
            d_sen[n_old:n_new, :] = sts_arr

        try:
            self.__hdf5_file.flush()
        except Exception:
            pass

        batch_count = sum(len(items) for items in buckets.values())
        self.__writer_cnt += batch_count
        if self.__writer_cnt % (self.__batch_size * 32) == 0:
            print("#" * 50)
            for group_name in self.__hdf5_file.keys():
                print(f"{group_name} len:{self.get_shape(group_name)[0]}")

    def create_dataset(
            self,
            group_name: str,
            shape: tuple,
            dtype: np.dtype,
            chunk_num: int,
            compression=hdf5plugin.Bitshuffle(nelems=0, cname='lz4'),
    ):
        if group_name not in self.__group_dict:
            self.__group_dict[group_name] = self.__hdf5_file.create_group(
                group_name)

        dataset = self.__group_dict[group_name].create_dataset(
            "data",
            shape=(0, *shape),
            maxshape=(None, *shape),
            dtype=dtype,
            chunks=(chunk_num, *shape),
            compression=compression,
        )
        get_ts_set = self.__group_dict[group_name].create_dataset(
            "get_ts",
            shape=(0, 1),
            maxshape=(None, 1),
            dtype=np.int64,
            chunks=(chunk_num, 1),
        )
        sen_ts_set = self.__group_dict[group_name].create_dataset(
            "sen_ts",
            shape=(0, 1),
            maxshape=(None, 1),
            dtype=np.int64,
            chunks=(chunk_num, 1),
        )

        # Store dataset reference for easy access
        self.__dataset_dict[f"{group_name}/data"] = dataset
        self.__dataset_dict[f"{group_name}/get_ts"] = get_ts_set
        self.__dataset_dict[f"{group_name}/sen_ts"] = sen_ts_set

    def append_data(
        self,
        group_name: str,
        data: np.ndarray,
        get_ts: np.ndarray,
        sen_ts: np.ndarray,
        block: bool = True,
        timeout: float = None,
    ):
        item = (
            group_name,
            data,
            get_ts,
            sen_ts,
        )
        self.__queue.put(item, block=block, timeout=timeout)

    def append_batch_data(
        self,
        group_name: str,
        data: np.ndarray,
        get_ts: np.ndarray,
        sen_ts: np.ndarray,
        block: bool = True,
        timeout: float = None,
    ):
        for i in range(data.shape[0]):
            self.append_data(
                group_name,
                data[i],
                get_ts[i],
                sen_ts[i],
                block=block,
                timeout=timeout,
            )

    def now_ns(self):
        return np.array([time.time_ns()])

    def hex_ts_to_ns(self, ts: dict):
        try:
            return np.array([ts["s"] * 1e9 + ts["ns"]])
        except Exception as e:
            print(f"hex_ts_to_ns failed: {e}")
            return np.array([np.inf])



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
################################################################
# Copyright 2025 Dong Zhaorui. All rights reserved.
# Author: Dong Zhaorui 847235539@qq.com
# Date  : 2025-09-18
################################################################

import copy
import numpy as np
from typing import Tuple


class HexObsUtilJoint:

    def __init__(
        self,
        mass: np.ndarray,
        damp: np.ndarray,
        stiff: np.ndarray,
        dt: float,
        q_limit: np.ndarray,
        dq_limit: np.ndarray,
        ddq_limit: np.ndarray,
    ):
        ### physical params
        self.__mass_inv = np.linalg.inv(mass)
        self.__damp = damp
        self.__stiff = stiff
        self.__dt = dt

        ### limits
        self.__q_limit = q_limit
        self.__dq_limit = dq_limit
        self.__ddq_limit = ddq_limit

        ### variables
        self.__ready = False
        self.__obs_q = None
        self.__obs_dq = None

    def get_mass(self) -> np.ndarray:
        return np.linalg.inv(self.__mass_inv)

    def set_mass(self, mass: np.ndarray):
        self.__mass_inv = np.linalg.inv(mass)

    def get_damp(self) -> np.ndarray:
        return copy.deepcopy(self.__damp)

    def set_damp(self, damp: np.ndarray):
        self.__damp = copy.deepcopy(damp)

    def get_stiff(self) -> np.ndarray:
        return copy.deepcopy(self.__stiff)

    def set_stiff(self, stiff: np.ndarray):
        self.__stiff = copy.deepcopy(stiff)

    def get_dt(self) -> float:
        return self.__dt

    def set_dt(self, dt: float):
        self.__dt = dt

    def is_ready(self) -> bool:
        return self.__ready

    def set_state(self, q: np.ndarray, dq: np.ndarray):
        self.__obs_q = q
        self.__obs_dq = dq
        self.__ready = True

    def get_state(self) -> Tuple[np.ndarray, np.ndarray]:
        return self.__obs_q, self.__obs_dq

    def predict(
        self,
        q_tar: np.ndarray,
    ):
        q_cur = self.__obs_q
        dq_cur = self.__obs_dq

        # runge-kutta k1
        dq1 = dq_cur
        ddq1 = self.__ddq(q_tar - q_cur, dq_cur)

        # runge-kutta k2
        q2 = q_cur + dq1 * self.__dt * 0.5
        dq2 = dq_cur + ddq1 * self.__dt * 0.5
        ddq2 = self.__ddq(q_tar - q2, dq2)

        # runge-kutta k3
        q3 = q_cur + dq2 * self.__dt * 0.5
        dq3 = dq_cur + ddq2 * self.__dt * 0.5
        ddq3 = self.__ddq(q_tar - q3, dq3)

        # runge-kutta k4
        q4 = q_cur + dq3 * self.__dt
        dq4 = dq_cur + ddq3 * self.__dt
        ddq4 = self.__ddq(q_tar - q4, dq4)

        # runge-kutta
        q_next = q_cur + (dq1 + 2.0 * dq2 + 2.0 * dq3 + dq4) / 6.0 * self.__dt
        dq_next = dq_cur + (ddq1 + 2.0 * ddq2 + 2.0 * ddq3 +
                            ddq4) / 6.0 * self.__dt
        low_mask = q_next < self.__q_limit[:, 0]
        high_mask = q_next > self.__q_limit[:, 1]

        # clip
        q_next[low_mask] = self.__q_limit[low_mask, 0]
        q_next[high_mask] = self.__q_limit[high_mask, 1]
        dq_next[low_mask] = 0.0
        dq_next[high_mask] = 0.0
        dq_next = np.clip(
            dq_next,
            self.__dq_limit[:, 0],
            self.__dq_limit[:, 1],
        )

        # set state
        self.__obs_q = q_next
        self.__obs_dq = dq_next

    def __ddq(
        self,
        q_err: np.ndarray,
        dq_cur: np.ndarray,
    ) -> np.ndarray:
        ddq = (self.__stiff @ q_err - self.__damp @ dq_cur) @ self.__mass_inv
        ddq = np.clip(ddq, self.__ddq_limit[:, 0], self.__ddq_limit[:, 1])
        return ddq

    def update(
        self,
        q_sensor: np.ndarray,
        dq_sensor: np.ndarray,
        weight_sensor: np.ndarray,
    ):
        q_sensor = np.clip(q_sensor, self.__q_limit[:, 0], self.__q_limit[:,
                                                                          1])
        dq_sensor = np.clip(dq_sensor, self.__dq_limit[:, 0],
                            self.__dq_limit[:, 1])

        # update state
        weight_intgr = 1.0 - weight_sensor
        self.__obs_q = self.__obs_q * weight_intgr + q_sensor * weight_sensor
        self.__obs_dq = self.__obs_dq * weight_intgr + dq_sensor * weight_sensor


class HexObsUtilDisturbance:

    def __init__(
            self,
            dt: float = 4e-3,
            fc: np.ndarray = np.array([10.0] * 6),
            fa: np.ndarray = np.array([25.0] * 6),
            dis_limit: np.ndarray = np.array([3.0] * 6),
    ):
        ### limits
        self.__dis_upper = dis_limit.copy()
        self.__dis_lower = -dis_limit.copy()

        ### constants
        self.__dt = dt
        self.__alpha_q = np.exp(-2.0 * np.pi * fc * dt)
        self.__alpha_a = np.exp(-2.0 * np.pi * fa * dt)

        ### variables
        self.__ready = False
        self.__dof = dis_limit.shape[0]
        self.__ddq_hat = np.zeros(self.__dof)
        self.__dis_hat = np.zeros(self.__dof)
        self.__last_cmd = np.zeros(self.__dof)
        self.__last_fric = np.zeros(self.__dof)

    def ready(self):
        return self.__ready

    def set_last(self, cmd: np.ndarray, fric: np.ndarray):
        self.__last_cmd = cmd.copy()
        self.__last_fric = fric.copy()
        self.__ready = True

    def __call__(
        self,
        dq: np.ndarray,
        ddq_raw: np.ndarray,
        m_mat: np.ndarray,
        c_mat: np.ndarray,
        g_vec: np.ndarray,
    ):
        # acc estimate
        self.__ddq_hat = self.__alpha_a * self.__ddq_hat + (
            1 - self.__alpha_a) * ddq_raw

        # nominal dynamics
        tau_model = m_mat @ self.__ddq_hat + c_mat @ dq + g_vec

        # raw disturbance calculation
        dis_raw = self.__last_cmd - tau_model - self.__last_fric
        self.__dis_hat = self.__alpha_q * self.__dis_hat + (
            1 - self.__alpha_q) * dis_raw
        self.__dis_hat = np.clip(self.__dis_hat, self.__dis_lower,
                                 self.__dis_upper)

        return self.__dis_hat



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
################################################################
# Copyright 2025 Dong Zhaorui. All rights reserved.
# Author: Dong Zhaorui 847235539@qq.com
# Date  : 2025-01-14
################################################################

# utils
from .dyn_util import HexDynUtil
from .dyn_util import HexFricUtil
from .obs_util import HexObsUtilJoint
from .obs_util import HexObsUtilDisturbance
from .ctrl_util import HexCtrlUtilMitJoint
from .ctrl_util import HexCtrlUtilPidJoint
from .ctrl_util import HexCtrlUtilIntJoint
from .ctrl_util import HexCtrlUtilMitWork
from .ctrl_util import HexCtrlUtilIntWork
from .hdf5_reader import HexHdf5Reader
from .hdf5_writer import HexHdf5Writer

# basic
from .math_utils import hat
from .math_utils import vee
from .math_utils import rad2deg
from .math_utils import deg2rad
from .math_utils import angle_norm
from .math_utils import quat_slerp
from .math_utils import quat_mul
from .math_utils import quat_inv
from .math_utils import trans_inv

# rotation
from .math_utils import rot2quat
from .math_utils import rot2axis
from .math_utils import rot2so3
from .math_utils import quat2rot
from .math_utils import quat2axis
from .math_utils import quat2so3
from .math_utils import axis2rot
from .math_utils import axis2quat
from .math_utils import axis2so3
from .math_utils import so32rot
from .math_utils import so32quat
from .math_utils import so32axis

# pose
from .math_utils import trans2part
from .math_utils import trans2se3
from .math_utils import part2trans
from .math_utils import part2se3
from .math_utils import se32trans
from .math_utils import se32part

# euler
from .math_utils import zyz2rot
from .math_utils import rot2zyz
from .math_utils import yaw2quat
from .math_utils import quat2yaw

__all__ = [
    # version
    '__version__',

    # utils
    'HexDynUtil',
    'HexFricUtil',
    'HexObsUtilJoint',
    'HexObsUtilDisturbance',
    'HexCtrlUtilMitJoint',
    'HexCtrlUtilPidJoint',
    'HexCtrlUtilIntJoint',
    'HexCtrlUtilMitWork',
    'HexCtrlUtilIntWork',
    'HexCtrlUtilPid',
    'HexCtrlUtilInt',
    'HexHdf5Reader',
    'HexHdf5Writer',

    # math basic
    'hat',
    'vee',
    'rad2deg',
    'deg2rad',
    'angle_norm',
    'quat_slerp',
    'quat_mul',
    'quat_inv',
    'trans_inv',

    # math rotation
    'rot2quat',
    'rot2axis',
    'rot2so3',
    'quat2rot',
    'quat2axis',
    'quat2so3',
    'axis2rot',
    'axis2quat',
    'axis2so3',
    'so32rot',
    'so32quat',
    'so32axis',

    # math pose
    'trans2part',
    'trans2se3',
    'part2trans',
    'part2se3',
    'se32trans',
    'se32part',

    # math euler
    'zyz2rot',
    'rot2zyz',
    'yaw2quat',
    'quat2yaw',
]

# print("#### Thanks for using HEXFELLOW Utilities :) ####")



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
################################################################
# Copyright 2025 Dong Zhaorui. All rights reserved.
# Author: Dong Zhaorui 847235539@qq.com
# Date  : 2025-01-14
################################################################

import numpy as np
from typing import Tuple


def hat(vec: np.ndarray):
    """so(3) vector → skew-symmetric matrix"""
    assert (vec.ndim == 1 and vec.shape == (3, )) or (
        vec.ndim == 2 and vec.shape[1] == 3), "cross_matrix vec shape err"

    if vec.ndim == 1:
        mat = np.array([
            [0.0, -vec[2], vec[1]],
            [vec[2], 0.0, -vec[0]],
            [-vec[1], vec[0], 0.0],
        ])
    else:
        # Batch processing: input shape (N, 3) -> output shape (N, 3, 3)
        num = vec.shape[0]
        mat = np.zeros((num, 3, 3))
        mat[:, 0, 1] = -vec[:, 2]
        mat[:, 0, 2] = vec[:, 1]
        mat[:, 1, 0] = vec[:, 2]
        mat[:, 1, 2] = -vec[:, 0]
        mat[:, 2, 0] = -vec[:, 1]
        mat[:, 2, 1] = vec[:, 0]
    return mat


def vee(mat: np.ndarray):
    """skew-symmetric matrix → so(3) vector"""
    assert (mat.ndim == 2 and mat.shape == (3, 3)) or (
        mat.ndim == 3 and mat.shape[1:] == (3, 3)), "vee mat shape err"

    if mat.ndim == 2:
        vec = np.array([mat[2, 1], mat[0, 2], mat[1, 0]])
    else:
        # Batch processing: input shape (N, 3, 3) -> output shape (N, 3)
        vec = np.zeros((mat.shape[0], 3))
        vec[:, 0] = mat[:, 2, 1]
        vec[:, 1] = mat[:, 0, 2]
        vec[:, 2] = mat[:, 1, 0]
    return vec


def rad2deg(rad):
    deg = rad * 180.0 / np.pi
    return deg


def deg2rad(deg):
    rad = deg * np.pi / 180.0
    return rad


def angle_norm(rad):
    normed_rad = (rad + np.pi) % (2 * np.pi) - np.pi
    return normed_rad


def quat_slerp(q1: np.ndarray, q2: np.ndarray, t) -> np.ndarray:
    assert ((q1.ndim == 1 and q1.shape == (4, ))
            or (q1.ndim == 2 and q1.shape[1] == 4)), "quat_slerp q1 shape err"
    assert ((q2.ndim == 1 and q2.shape == (4, ))
            or (q2.ndim == 2 and q2.shape[1] == 4)), "quat_slerp q2 shape err"
    assert q1.ndim == q2.ndim, "quat_slerp q1 and q2 must have same ndim"
    if q1.ndim == 2:
        assert q1.shape[0] == q2.shape[0], "quat_slerp batch size mismatch"
        assert (isinstance(t, np.ndarray) and t.ndim == 1 and
                t.shape[0] == q1.shape[0]) or isinstance(t, (int, float)), \
            "quat_slerp t shape err"

    if q1.ndim == 1:
        assert isinstance(
            t, (int, float)), "quat_slerp t must be scalar for 1D quat"
        # normalize
        q1_norm = q1 / np.linalg.norm(q1)
        if q1_norm[0] < 0.0:
            q1_norm = -q1_norm
        q2_norm = q2 / np.linalg.norm(q2)
        if q2_norm[0] < 0.0:
            q2_norm = -q2_norm

        # dot
        dot = np.dot(q1_norm, q2_norm)
        if dot < 0.0:
            q2_norm = -q2_norm
            dot = -dot
        dot = np.clip(dot, -1.0, 1.0)
        theta = np.arccos(dot)

        # slerp
        if np.fabs(theta) < 1e-6:
            q = q1_norm + t * (q2_norm - q1_norm)
            q = q / np.linalg.norm(q)
            return q

        sin_theta = np.sin(theta)
        q1_factor = np.sin((1 - t) * theta) / sin_theta
        q2_factor = np.sin(t * theta) / sin_theta
        q = q1_factor * q1_norm + q2_factor * q2_norm
        return q
    else:
        # Batch processing: input shape (N, 4), (N, 4), (N,) or scalar -> output shape (N, 4)
        if isinstance(t, (int, float)):
            t = np.full(q1.shape[0], t)

        # normalize
        q1_norm = q1 / np.linalg.norm(q1, axis=1, keepdims=True)
        neg_mask1 = q1_norm[:, 0] < 0.0
        q1_norm[neg_mask1] = -q1_norm[neg_mask1]
        q2_norm = q2 / np.linalg.norm(q2, axis=1, keepdims=True)
        neg_mask2 = q2_norm[:, 0] < 0.0
        q2_norm[neg_mask2] = -q2_norm[neg_mask2]

        # dot
        dot = np.sum(q1_norm * q2_norm, axis=1)
        neg_dot_mask = dot < 0.0
        q2_norm[neg_dot_mask] = -q2_norm[neg_dot_mask]
        dot[neg_dot_mask] = -dot[neg_dot_mask]
        dot = np.clip(dot, -1.0, 1.0)
        theta = np.arccos(dot)

        # slerp
        small_mask = np.fabs(theta) < 1e-6
        q = np.zeros_like(q1)

        if np.any(small_mask):
            q[small_mask] = q1_norm[small_mask] + t[small_mask, np.newaxis] * (
                q2_norm[small_mask] - q1_norm[small_mask])
            q[small_mask] = q[small_mask] / np.linalg.norm(
                q[small_mask], axis=1, keepdims=True)

        if np.any(~small_mask):
            sin_theta = np.sin(theta[~small_mask])
            q1_factor = np.sin(
                (1 - t[~small_mask]) * theta[~small_mask]) / sin_theta
            q2_factor = np.sin(t[~small_mask] * theta[~small_mask]) / sin_theta
            q[~small_mask] = (q1_factor[:, np.newaxis] * q1_norm[~small_mask] +
                              q2_factor[:, np.newaxis] * q2_norm[~small_mask])
        return q


def quat_mul(q1: np.ndarray, q2: np.ndarray) -> np.ndarray:
    assert ((q1.ndim == 1 and q1.shape == (4, ))
            or (q1.ndim == 2 and q1.shape[1] == 4)), "quat_mul q1 shape err"
    assert ((q2.ndim == 1 and q2.shape == (4, ))
            or (q2.ndim == 2 and q2.shape[1] == 4)), "quat_mul q2 shape err"
    assert q1.ndim == q2.ndim, "quat_mul q1 and q2 must have same ndim"
    if q1.ndim == 2:
        assert q1.shape[0] == q2.shape[0], "quat_mul batch size mismatch"

    # normalize
    if q1.ndim == 1:
        q1_norm = q1 / np.linalg.norm(q1)
        if q1_norm[0] < 0.0:
            q1_norm = -q1_norm
        q2_norm = q2 / np.linalg.norm(q2)
        if q2_norm[0] < 0.0:
            q2_norm = -q2_norm

        # mul
        w1, x1, y1, z1 = q1_norm
        w2, x2, y2, z2 = q2_norm
        w = w1 * w2 - x1 * x2 - y1 * y2 - z1 * z2
        x = w1 * x2 + x1 * w2 + y1 * z2 - z1 * y2
        y = w1 * y2 - x1 * z2 + y1 * w2 + z1 * x2
        z = w1 * z2 + x1 * y2 - y1 * x2 + z1 * w2
        q = np.array([w, x, y, z])
    else:
        # Batch processing
        q1_norm = q1 / np.linalg.norm(q1, axis=1, keepdims=True)
        neg_mask1 = q1_norm[:, 0] < 0.0
        q1_norm[neg_mask1] = -q1_norm[neg_mask1]
        q2_norm = q2 / np.linalg.norm(q2, axis=1, keepdims=True)
        neg_mask2 = q2_norm[:, 0] < 0.0
        q2_norm[neg_mask2] = -q2_norm[neg_mask2]

        # mul
        w1, x1, y1, z1 = q1_norm[:, 0], q1_norm[:, 1], q1_norm[:,
                                                               2], q1_norm[:,
                                                                           3]
        w2, x2, y2, z2 = q2_norm[:, 0], q2_norm[:, 1], q2_norm[:,
                                                               2], q2_norm[:,
                                                                           3]
        w = w1 * w2 - x1 * x2 - y1 * y2 - z1 * z2
        x = w1 * x2 + x1 * w2 + y1 * z2 - z1 * y2
        y = w1 * y2 - x1 * z2 + y1 * w2 + z1 * x2
        z = w1 * z2 + x1 * y2 - y1 * x2 + z1 * w2
        q = np.stack([w, x, y, z], axis=1)
    return q


def quat_inv(quat: np.ndarray) -> np.ndarray:
    assert (quat.ndim == 1 and quat.shape == (4, )) or (
        quat.ndim == 2 and quat.shape[1] == 4), "quat_inv quat shape err"

    if quat.ndim == 1:
        q = quat / np.linalg.norm(quat)
        if q[0] < 0.0:
            q = -q
        # inv
        inv = np.array([q[0], -q[1], -q[2], -q[3]])
    else:
        # Batch processing: input shape (N, 4) -> output shape (N, 4)
        q = quat / np.linalg.norm(quat, axis=1, keepdims=True)
        neg_mask = q[:, 0] < 0.0
        q[neg_mask] = -q[neg_mask]
        # inv
        inv = np.stack([q[:, 0], -q[:, 1], -q[:, 2], -q[:, 3]], axis=1)
    return inv


def trans_inv(trans: np.ndarray) -> np.ndarray:
    assert (trans.ndim == 2 and trans.shape == (4, 4)) or (
        trans.ndim == 3
        and trans.shape[1:] == (4, 4)), "trans_inv trans shape err"

    if trans.ndim == 2:
        pos = trans[:3, 3]
        rot = trans[:3, :3]

        # inv
        inv = np.eye(4)
        inv[:3, :3] = rot.T
        inv[:3, 3] = -inv[:3, :3] @ pos
        return inv
    else:
        # Batch processing: input shape (N, 4, 4) -> output shape (N, 4, 4)
        pos = trans[:, :3, 3]
        rot = trans[:, :3, :3]

        # inv
        inv = np.zeros_like(trans)
        inv[:, :3, :3] = rot.transpose(0, 2, 1)
        inv[:, :3, 3] = -np.einsum('ijk,ik->ij', inv[:, :3, :3], pos)
        inv[:, 3, :3] = 0.0
        inv[:, 3, 3] = 1.0
        return inv


def rot2quat(rot: np.ndarray) -> np.ndarray:
    assert (rot.ndim == 2 and rot.shape == (3, 3)) or (
        rot.ndim == 3 and rot.shape[1:] == (3, 3)), "rot2quat rot shape err"

    if rot.ndim == 2:
        qw, qx, qy, qz = 1, 0, 0, 0
        trace = np.trace(rot)
        if trace > 0:
            temp = 2.0 * np.sqrt(1 + trace)
            qw = 0.25 * temp
            qx = (rot[2, 1] - rot[1, 2]) / temp
            qy = (rot[0, 2] - rot[2, 0]) / temp
            qz = (rot[1, 0] - rot[0, 1]) / temp
        else:
            if rot[0, 0] > rot[1, 1] and rot[0, 0] > rot[2, 2]:
                temp = 2.0 * np.sqrt(1 + rot[0, 0] - rot[1, 1] - rot[2, 2])
                qw = (rot[2, 1] - rot[1, 2]) / temp
                qx = 0.25 * temp
                qy = (rot[1, 0] + rot[0, 1]) / temp
                qz = (rot[0, 2] + rot[2, 0]) / temp
            elif rot[1, 1] > rot[2, 2]:
                temp = 2.0 * np.sqrt(1 + rot[1, 1] - rot[0, 0] - rot[2, 2])
                qw = (rot[0, 2] - rot[2, 0]) / temp
                qx = (rot[1, 0] + rot[0, 1]) / temp
                qy = 0.25 * temp
                qz = (rot[2, 1] + rot[1, 2]) / temp
            else:
                temp = 2.0 * np.sqrt(1 + rot[2, 2] - rot[0, 0] - rot[1, 1])
                qw = (rot[1, 0] - rot[0, 1]) / temp
                qx = (rot[0, 2] + rot[2, 0]) / temp
                qy = (rot[2, 1] + rot[1, 2]) / temp
                qz = 0.25 * temp

        return np.array([qw, qx, qy, qz])
    else:
        # Batch processing: input shape (N, 3, 3) -> output shape (N, 4)
        num = rot.shape[0]
        quat = np.zeros((num, 4))
        trace = np.trace(rot, axis1=1, axis2=2)

        # Case 1: trace > 0
        mask1 = trace > 0
        if np.any(mask1):
            temp = 2.0 * np.sqrt(1 + trace[mask1])
            quat[mask1, 0] = 0.25 * temp
            quat[mask1, 1] = (rot[mask1, 2, 1] - rot[mask1, 1, 2]) / temp
            quat[mask1, 2] = (rot[mask1, 0, 2] - rot[mask1, 2, 0]) / temp
            quat[mask1, 3] = (rot[mask1, 1, 0] - rot[mask1, 0, 1]) / temp

        # Case 2: trace <= 0
        mask2 = ~mask1
        if np.any(mask2):
            rot_masked = rot[mask2]
            diag = np.diagonal(rot_masked, axis1=1, axis2=2)
            mask2a = (diag[:, 0] > diag[:, 1]) & (diag[:, 0] > diag[:, 2])
            mask2b = (~mask2a) & (diag[:, 1] > diag[:, 2])
            mask2c = (~mask2a) & (~mask2b)

            # Case 2a: rot[0,0] > rot[1,1] and rot[0,0] > rot[2,2]
            if np.any(mask2a):
                idx2a = np.where(mask2)[0][mask2a]
                temp = 2.0 * np.sqrt(1 + diag[mask2a, 0] - diag[mask2a, 1] -
                                     diag[mask2a, 2])
                quat[idx2a, 0] = (rot_masked[mask2a, 2, 1] -
                                  rot_masked[mask2a, 1, 2]) / temp
                quat[idx2a, 1] = 0.25 * temp
                quat[idx2a, 2] = (rot_masked[mask2a, 1, 0] +
                                  rot_masked[mask2a, 0, 1]) / temp
                quat[idx2a, 3] = (rot_masked[mask2a, 0, 2] +
                                  rot_masked[mask2a, 2, 0]) / temp

            # Case 2b: rot[1,1] > rot[2,2]
            if np.any(mask2b):
                idx2b = np.where(mask2)[0][mask2b]
                temp = 2.0 * np.sqrt(1 + diag[mask2b, 1] - diag[mask2b, 0] -
                                     diag[mask2b, 2])
                quat[idx2b, 0] = (rot_masked[mask2b, 0, 2] -
                                  rot_masked[mask2b, 2, 0]) / temp
                quat[idx2b, 1] = (rot_masked[mask2b, 1, 0] +
                                  rot_masked[mask2b, 0, 1]) / temp
                quat[idx2b, 2] = 0.25 * temp
                quat[idx2b, 3] = (rot_masked[mask2b, 2, 1] +
                                  rot_masked[mask2b, 1, 2]) / temp

            # Case 2c: rot[2,2] >= rot[0,0] and rot[2,2] >= rot[1,1]
            if np.any(mask2c):
                idx2c = np.where(mask2)[0][mask2c]
                temp = 2.0 * np.sqrt(1 + diag[mask2c, 2] - diag[mask2c, 0] -
                                     diag[mask2c, 1])
                quat[idx2c, 0] = (rot_masked[mask2c, 1, 0] -
                                  rot_masked[mask2c, 0, 1]) / temp
                quat[idx2c, 1] = (rot_masked[mask2c, 0, 2] +
                                  rot_masked[mask2c, 2, 0]) / temp
                quat[idx2c, 2] = (rot_masked[mask2c, 2, 1] +
                                  rot_masked[mask2c, 1, 2]) / temp
                quat[idx2c, 3] = 0.25 * temp

        return quat


def rot2axis(rot: np.ndarray) -> Tuple[np.ndarray, float]:
    assert (rot.ndim == 2 and rot.shape == (3, 3)) or (
        rot.ndim == 3 and rot.shape[1:] == (3, 3)), "rot2axis rot shape err"

    if rot.ndim == 2:
        cos_theta = 0.5 * (np.trace(rot) - 1)
        cos_theta = np.clip(cos_theta, -1.0, 1.0)
        theta = np.arccos(cos_theta)

        if theta < 1e-6:
            return np.array([1.0, 0.0, 0.0]), 0.0
        else:
            axis_matrix = (rot - rot.T) / (2 * np.sin(theta))
            axis = vee(axis_matrix)
        return axis, theta
    else:
        # Batch processing: input shape (N, 3, 3) -> output (N, 3), (N,)
        cos_theta = 0.5 * (np.trace(rot, axis1=1, axis2=2) - 1)
        cos_theta = np.clip(cos_theta, -1.0, 1.0)
        theta = np.arccos(cos_theta)

        small_mask = theta < 1e-6
        axis = np.zeros((rot.shape[0], 3))
        axis[small_mask] = np.array([1.0, 0.0, 0.0])

        if np.any(~small_mask):
            axis_matrix = (rot[~small_mask] -
                           rot[~small_mask].transpose(0, 2, 1)) / (2 * np.sin(
                               theta[~small_mask, np.newaxis, np.newaxis]))
            axis[~small_mask] = vee(axis_matrix)

        return axis, theta


def rot2so3(rot: np.ndarray) -> np.ndarray:
    assert (rot.ndim == 2 and rot.shape == (3, 3)) or (
        rot.ndim == 3 and rot.shape[1:] == (3, 3)), "rot2so3 rot shape err"

    axis, theta = rot2axis(rot)
    if rot.ndim == 2:
        return theta * axis
    else:
        # Batch processing: theta is (N,), axis is (N, 3)
        return theta[:, np.newaxis] * axis


def quat2rot(quat: np.ndarray) -> np.ndarray:
    assert (quat.ndim == 1 and quat.shape == (4, )) or (
        quat.ndim == 2 and quat.shape[1] == 4), "quat2rot quat shape err"

    if quat.ndim == 1:
        q = quat / np.linalg.norm(quat)
        if q[0] < 0.0:
            q = -q

        # temp vars
        qx2 = q[1] * q[1]
        qy2 = q[2] * q[2]
        qz2 = q[3] * q[3]
        qxqw = q[1] * q[0]
        qyqw = q[2] * q[0]
        qzqw = q[3] * q[0]
        qxqy = q[1] * q[2]
        qyqz = q[2] * q[3]
        qzqx = q[3] * q[1]

        # rot
        rot = np.array([
            [
                1 - 2 * (qy2 + qz2),
                2 * (qxqy - qzqw),
                2 * (qzqx + qyqw),
            ],
            [
                2 * (qxqy + qzqw),
                1 - 2 * (qx2 + qz2),
                2 * (qyqz - qxqw),
            ],
            [
                2 * (qzqx - qyqw),
                2 * (qyqz + qxqw),
                1 - 2 * (qx2 + qy2),
            ],
        ])
    else:
        # Batch processing: input shape (N, 4) -> output shape (N, 3, 3)
        q = quat / np.linalg.norm(quat, axis=1, keepdims=True)
        neg_mask = q[:, 0] < 0.0
        q[neg_mask] = -q[neg_mask]

        # temp vars
        qx2 = q[:, 1] * q[:, 1]
        qy2 = q[:, 2] * q[:, 2]
        qz2 = q[:, 3] * q[:, 3]
        qxqw = q[:, 1] * q[:, 0]
        qyqw = q[:, 2] * q[:, 0]
        qzqw = q[:, 3] * q[:, 0]
        qxqy = q[:, 1] * q[:, 2]
        qyqz = q[:, 2] * q[:, 3]
        qzqx = q[:, 3] * q[:, 1]

        # rot
        num = q.shape[0]
        rot = np.zeros((num, 3, 3))
        rot[:, 0, 0] = 1 - 2 * (qy2 + qz2)
        rot[:, 0, 1] = 2 * (qxqy - qzqw)
        rot[:, 0, 2] = 2 * (qzqx + qyqw)
        rot[:, 1, 0] = 2 * (qxqy + qzqw)
        rot[:, 1, 1] = 1 - 2 * (qx2 + qz2)
        rot[:, 1, 2] = 2 * (qyqz - qxqw)
        rot[:, 2, 0] = 2 * (qzqx - qyqw)
        rot[:, 2, 1] = 2 * (qyqz + qxqw)
        rot[:, 2, 2] = 1 - 2 * (qx2 + qy2)
    return rot


def quat2axis(quat: np.ndarray) -> Tuple[np.ndarray, float]:
    assert (quat.ndim == 1 and quat.shape == (4, )) or (
        quat.ndim == 2 and quat.shape[1] == 4), "quat2axis quat shape err"

    if quat.ndim == 1:
        q = quat / np.linalg.norm(quat)
        if q[0] < 0.0:
            q = -q

        vec = q[1:]
        norm_vec = np.linalg.norm(vec)
        if norm_vec < 1e-6:
            return np.array([1.0, 0.0, 0.0]), 0.0

        theta = 2 * np.arctan2(norm_vec, q[0])
        axis = vec / norm_vec
        return axis, theta
    else:
        # Batch processing: input shape (N, 4) -> output (N, 3), (N,)
        q = quat / np.linalg.norm(quat, axis=1, keepdims=True)
        neg_mask = q[:, 0] < 0.0
        q[neg_mask] = -q[neg_mask]

        vec = q[:, 1:]
        norm_vec = np.linalg.norm(vec, axis=1)
        small_mask = norm_vec < 1e-6

        axis = np.zeros((q.shape[0], 3))
        axis[~small_mask] = vec[~small_mask] / norm_vec[~small_mask,
                                                        np.newaxis]
        axis[small_mask] = np.array([1.0, 0.0, 0.0])

        theta = np.zeros(q.shape[0])
        theta[~small_mask] = 2 * np.arctan2(norm_vec[~small_mask],
                                            q[~small_mask, 0])
        return axis, theta


def quat2so3(quat: np.ndarray) -> np.ndarray:
    assert (quat.ndim == 1 and quat.shape == (4, )) or (
        quat.ndim == 2 and quat.shape[1] == 4), "quat2so3 quat shape err"

    axis, theta = quat2axis(quat)
    if quat.ndim == 1:
        return theta * axis
    else:
        # Batch processing: theta is (N,), axis is (N, 3)
        return theta[:, np.newaxis] * axis


def axis2rot(axis: np.ndarray, theta) -> np.ndarray:
    assert (axis.ndim == 1 and axis.shape == (3, )) or (
        axis.ndim == 2 and axis.shape[1] == 3), "axis2rot axis shape err"
    if axis.ndim == 1:
        assert isinstance(
            theta, (int, float)), "axis2rot theta must be scalar for 1D axis"
    else:
        assert (isinstance(theta, np.ndarray) and theta.ndim == 1 and
                theta.shape[0] == axis.shape[0]) or isinstance(theta, (int, float)), \
            "axis2rot theta shape err"

    if axis.ndim == 1:
        if theta < 1e-6:
            return np.eye(3)

        axis_matrix = hat(axis)
        rot = np.eye(3) + np.sin(theta) * axis_matrix + (1 - np.cos(theta)) * (
            axis_matrix @ axis_matrix)
        return rot
    else:
        # Batch processing: input shape (N, 3), (N,) or scalar -> output shape (N, 3, 3)
        if isinstance(theta, (int, float)):
            theta = np.full(axis.shape[0], theta)

        num = axis.shape[0]
        rot = np.zeros((num, 3, 3))
        small_mask = theta < 1e-6
        rot[small_mask] = np.eye(3)

        if np.any(~small_mask):
            axis_matrix = hat(axis[~small_mask])
            sin_theta = np.sin(theta[~small_mask])
            cos_theta = np.cos(theta[~small_mask])
            # Batch matrix multiplication
            axis_matrix_sq = axis_matrix @ axis_matrix
            rot[~small_mask] = (
                np.eye(3)[np.newaxis, :, :] +
                sin_theta[:, np.newaxis, np.newaxis] * axis_matrix +
                (1 - cos_theta)[:, np.newaxis, np.newaxis] * axis_matrix_sq)
        return rot


def axis2quat(axis: np.ndarray, theta) -> np.ndarray:
    assert (axis.ndim == 1 and axis.shape == (3, )) or (
        axis.ndim == 2 and axis.shape[1] == 3), "axis2quat axis shape err"
    if axis.ndim == 1:
        assert isinstance(
            theta, (int, float)), "axis2quat theta must be scalar for 1D axis"
    else:
        assert (isinstance(theta, np.ndarray) and theta.ndim == 1 and
                theta.shape[0] == axis.shape[0]) or isinstance(theta, (int, float)), \
            "axis2quat theta shape err"

    if axis.ndim == 1:
        if theta < 1e-6:
            return np.array([1.0, 0.0, 0.0, 0.0])

        quat = np.zeros(4)
        quat[0] = np.cos(theta / 2)
        quat[1:] = axis * np.sin(theta / 2)
        return quat
    else:
        # Batch processing: input shape (N, 3), (N,) or scalar -> output shape (N, 4)
        if isinstance(theta, (int, float)):
            theta = np.full(axis.shape[0], theta)

        num = axis.shape[0]
        quat = np.zeros((num, 4))
        small_mask = theta < 1e-6
        quat[small_mask] = np.array([1.0, 0.0, 0.0, 0.0])

        if np.any(~small_mask):
            half_theta = theta[~small_mask] / 2
            quat[~small_mask, 0] = np.cos(half_theta)
            quat[~small_mask,
                 1:] = axis[~small_mask] * np.sin(half_theta)[:, np.newaxis]
        return quat


def axis2so3(axis: np.ndarray, theta) -> np.ndarray:
    assert (axis.ndim == 1 and axis.shape == (3, )) or (
        axis.ndim == 2 and axis.shape[1] == 3), "axis2so3 axis shape err"
    if axis.ndim == 1:
        assert isinstance(
            theta, (int, float)), "axis2so3 theta must be scalar for 1D axis"
    else:
        assert (isinstance(theta, np.ndarray) and theta.ndim == 1 and
                theta.shape[0] == axis.shape[0]) or isinstance(theta, (int, float)), \
            "axis2so3 theta shape err"

    if axis.ndim == 1:
        return theta * axis
    else:
        # Batch processing: input shape (N, 3), (N,) or scalar -> output shape (N, 3)
        if isinstance(theta, (int, float)):
            theta = np.full(axis.shape[0], theta)
        return theta[:, np.newaxis] * axis


def so32rot(so3: np.ndarray) -> np.ndarray:
    assert (so3.ndim == 1 and so3.shape == (3, )) or (
        so3.ndim == 2 and so3.shape[1] == 3), "so32rot so3 shape err"

    if so3.ndim == 1:
        theta = np.linalg.norm(so3)
        if theta < 1e-6:
            return np.eye(3)
        else:
            axis = so3 / theta
            return axis2rot(axis, theta)
    else:
        # Batch processing: input shape (N, 3) -> output shape (N, 3, 3)
        theta = np.linalg.norm(so3, axis=1)
        axis = so3 / theta[:, np.newaxis]
        return axis2rot(axis, theta)


def so32quat(so3: np.ndarray) -> np.ndarray:
    assert (so3.ndim == 1 and so3.shape == (3, )) or (
        so3.ndim == 2 and so3.shape[1] == 3), "so32quat so3 shape err"

    axis, theta = so32axis(so3)
    return axis2quat(axis, theta)


def so32axis(so3: np.ndarray) -> Tuple[np.ndarray, float]:
    assert (so3.ndim == 1 and so3.shape == (3, )) or (
        so3.ndim == 2 and so3.shape[1] == 3), "so32axis so3 shape err"

    if so3.ndim == 1:
        theta = np.linalg.norm(so3)
        if theta < 1e-6:
            return np.array([1.0, 0.0, 0.0]), 0.0
        else:
            axis = so3 / theta
            return axis, theta
    else:
        # Batch processing: input shape (N, 3) -> output (N, 3), (N,)
        theta = np.linalg.norm(so3, axis=1)
        small_mask = theta < 1e-6
        axis = np.zeros((so3.shape[0], 3))
        axis[small_mask] = np.array([1.0, 0.0, 0.0])
        axis[~small_mask] = so3[~small_mask] / theta[~small_mask, np.newaxis]
        return axis, theta


def trans2part(trans: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    assert (trans.ndim == 2 and trans.shape == (4, 4)) or (
        trans.ndim == 3
        and trans.shape[1:] == (4, 4)), "trans2part trans shape err"

    if trans.ndim == 2:
        pos = trans[:3, 3]
        quat = rot2quat(trans[:3, :3])
        return pos, quat
    else:
        # Batch processing: input shape (N, 4, 4) -> output (N, 3), (N, 4)
        pos = trans[:, :3, 3]
        quat = rot2quat(trans[:, :3, :3])
        return pos, quat


def trans2se3(trans: np.ndarray) -> np.ndarray:
    assert (trans.ndim == 2 and trans.shape == (4, 4)) or (
        trans.ndim == 3
        and trans.shape[1:] == (4, 4)), "trans2se3 trans shape err"

    if trans.ndim == 2:
        return np.concatenate((trans[:3, 3], rot2so3(trans[:3, :3])))
    else:
        # Batch processing: input shape (N, 4, 4) -> output shape (N, 6)
        pos = trans[:, :3, 3]
        so3 = rot2so3(trans[:, :3, :3])
        return np.concatenate((pos, so3), axis=1)


def part2trans(pos: np.ndarray, quat: np.ndarray) -> np.ndarray:
    assert (pos.ndim == 1 and pos.shape == (3, )) or (
        pos.ndim == 2 and pos.shape[1] == 3), "part2trans pos shape err"
    assert (quat.ndim == 1 and quat.shape == (4, )) or (
        quat.ndim == 2 and quat.shape[1] == 4), "part2trans quat shape err"
    assert pos.ndim == quat.ndim, "part2trans pos and quat must have same ndim"
    if pos.ndim == 2:
        assert pos.shape[0] == quat.shape[0], "part2trans batch size mismatch"

    if pos.ndim == 1:
        trans = np.eye(4)
        trans[:3, 3] = pos
        trans[:3, :3] = quat2rot(quat)
        return trans
    else:
        # Batch processing: input shape (N, 3), (N, 4) -> output shape (N, 4, 4)
        num = pos.shape[0]
        trans = np.zeros((num, 4, 4))
        trans[:, :3, 3] = pos
        trans[:, :3, :3] = quat2rot(quat)
        trans[:, 3, :3] = 0.0
        trans[:, 3, 3] = 1.0
        return trans


def part2se3(pos: np.ndarray, quat: np.ndarray) -> np.ndarray:
    assert (pos.ndim == 1 and pos.shape == (3, )) or (
        pos.ndim == 2 and pos.shape[1] == 3), "part2se3 pos shape err"
    assert (quat.ndim == 1 and quat.shape == (4, )) or (
        quat.ndim == 2 and quat.shape[1] == 4), "part2se3 quat shape err"
    assert pos.ndim == quat.ndim, "part2se3 pos and quat must have same ndim"
    if pos.ndim == 2:
        assert pos.shape[0] == quat.shape[0], "part2se3 batch size mismatch"

    if pos.ndim == 1:
        se3 = np.concatenate((pos, quat2so3(quat)))
        return se3
    else:
        # Batch processing: input shape (N, 3), (N, 4) -> output shape (N, 6)
        so3 = quat2so3(quat)
        return np.concatenate((pos, so3), axis=1)


def se32trans(se3: np.ndarray) -> np.ndarray:
    assert (se3.ndim == 1 and se3.shape == (6, )) or (
        se3.ndim == 2 and se3.shape[1] == 6), "se32trans se3 shape err"

    if se3.ndim == 1:
        trans = np.eye(4)
        trans[:3, 3] = se3[:3]
        trans[:3, :3] = so32rot(se3[3:])
        return trans
    else:
        # Batch processing: input shape (N, 6) -> output shape (N, 4, 4)
        num = se3.shape[0]
        trans = np.zeros((num, 4, 4))
        trans[:, :3, 3] = se3[:, :3]
        trans[:, :3, :3] = so32rot(se3[:, 3:])
        trans[:, 3, :3] = 0.0
        trans[:, 3, 3] = 1.0
        return trans


def se32part(se3: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    assert (se3.ndim == 1 and se3.shape == (6, )) or (
        se3.ndim == 2 and se3.shape[1] == 6), "se32part se3 shape err"

    if se3.ndim == 1:
        return se3[:3], so32quat(se3[3:])
    else:
        # Batch processing: input shape (N, 6) -> output (N, 3), (N, 4)
        return se3[:, :3], so32quat(se3[:, 3:])


def zyz2rot(zyz: np.ndarray) -> np.ndarray:
    assert (zyz.ndim == 1 and zyz.shape == (3, )) or (
        zyz.ndim == 2 and zyz.shape[1] == 3), "zyz2rot zyz shape err"

    if zyz.ndim == 1:
        # temp vars
        theta1, theta2, theta3 = zyz
        cos1 = np.cos(theta1)
        sin1 = np.sin(theta1)
        cos2 = np.cos(theta2)
        sin2 = np.sin(theta2)
        cos3 = np.cos(theta3)
        sin3 = np.sin(theta3)

        # rot
        rot = np.array([
            [
                cos1 * cos2 * cos3 - sin1 * sin3,
                -cos1 * cos2 * sin3 - sin1 * cos3,
                cos1 * sin2,
            ],
            [
                sin1 * cos2 * cos3 + cos1 * sin3,
                -sin1 * cos2 * sin3 + cos1 * cos3,
                sin1 * sin2,
            ],
            [
                -sin2 * cos3,
                sin2 * sin3,
                cos2,
            ],
        ])
        return rot
    else:
        # Batch processing: input shape (N, 3) -> output shape (N, 3, 3)
        theta1, theta2, theta3 = zyz[:, 0], zyz[:, 1], zyz[:, 2]
        cos1 = np.cos(theta1)
        sin1 = np.sin(theta1)
        cos2 = np.cos(theta2)
        sin2 = np.sin(theta2)
        cos3 = np.cos(theta3)
        sin3 = np.sin(theta3)

        # rot
        num = zyz.shape[0]
        rot = np.zeros((num, 3, 3))
        rot[:, 0, 0] = cos1 * cos2 * cos3 - sin1 * sin3
        rot[:, 0, 1] = -cos1 * cos2 * sin3 - sin1 * cos3
        rot[:, 0, 2] = cos1 * sin2
        rot[:, 1, 0] = sin1 * cos2 * cos3 + cos1 * sin3
        rot[:, 1, 1] = -sin1 * cos2 * sin3 + cos1 * cos3
        rot[:, 1, 2] = sin1 * sin2
        rot[:, 2, 0] = -sin2 * cos3
        rot[:, 2, 1] = sin2 * sin3
        rot[:, 2, 2] = cos2
        return rot


def rot2zyz(rot: np.ndarray, neg: bool = False) -> np.ndarray:
    assert (rot.ndim == 2 and rot.shape == (3, 3)) or (
        rot.ndim == 3 and rot.shape[1:] == (3, 3)), "rot2zyz rot shape err"

    if rot.ndim == 2:
        # positive
        theta1 = np.arctan2(rot[1, 2], rot[0, 2])
        theta2 = np.arccos(rot[2, 2])
        theta3 = np.arctan2(rot[2, 1], -rot[2, 0])

        # negative
        if neg:
            theta1 = angle_norm(theta1 + np.pi)
            theta2 = -theta2
            theta3 = angle_norm(theta3 + np.pi)

        return np.array([theta1, theta2, theta3])
    else:
        # Batch processing: input shape (N, 3, 3) -> output shape (N, 3)
        # positive
        theta1 = np.arctan2(rot[:, 1, 2], rot[:, 0, 2])
        theta2 = np.arccos(rot[:, 2, 2])
        theta3 = np.arctan2(rot[:, 2, 1], -rot[:, 2, 0])

        # negative
        if neg:
            theta1 = angle_norm(theta1 + np.pi)
            theta2 = -theta2
            theta3 = angle_norm(theta3 + np.pi)

        return np.stack([theta1, theta2, theta3], axis=1)


def yaw2quat(yaw) -> np.ndarray:
    if isinstance(yaw, (int, float)):
        quat = np.array([np.cos(yaw / 2), 0, 0, np.sin(yaw / 2)])
        return quat
    else:
        # Batch processing: input shape (N,) -> output shape (N, 4)
        assert isinstance(
            yaw, np.ndarray) and yaw.ndim == 1, "yaw2quat yaw shape err"
        quat = np.zeros((yaw.shape[0], 4))
        quat[:, 0] = np.cos(yaw / 2)
        quat[:, 3] = np.sin(yaw / 2)
        return quat


def quat2yaw(quat: np.ndarray):
    assert (quat.ndim == 1 and quat.shape == (4, )) or (
        quat.ndim == 2 and quat.shape[1] == 4), "quat2yaw quat shape err"

    if quat.ndim == 1:
        yaw = 2 * np.arctan2(quat[3], quat[0])
        return yaw
    else:
        # Batch processing: input shape (N, 4) -> output shape (N,)
        yaw = 2 * np.arctan2(quat[:, 3], quat[:, 0])
        return yaw



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
################################################################
# Copyright 2025 Dong Zhaorui. All rights reserved.
# Author: Dong Zhaorui 847235539@qq.com
# Date  : 2025-09-18
################################################################

import copy
import numpy as np
import pinocchio as pin
from typing import Tuple, List

from hex_robo_utils.math_utils import trans2part, part2trans
from hex_robo_utils.math_utils import trans_inv, trans2se3
from hex_robo_utils.math_utils import angle_norm, hat


class HexDynUtil:

    def __init__(
            self,
            model_path: str,
            last_link: str,
            end_pose: np.ndarray = np.array(
                [0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0]),
            gravity: np.ndarray = np.array([0, 0, -9.81]),
    ):
        ### pinocchio init
        self.__model = pin.buildModelFromUrdf(model_path)
        self.__data = self.__model.createData()
        self.__joint_num = self.__model.njoints - 1
        self.__end_link_id = self.__model.getFrameId(last_link)
        self.__end_joint_id = self.__joint_num
        self.__lower_limit = self.__model.lowerPositionLimit
        self.__upper_limit = self.__model.upperPositionLimit
        self.__trans_end_in_last = part2trans(end_pose[:3], end_pose[3:])
        self.__trans_last_in_end = trans_inv(self.__trans_end_in_last)
        self.__jac_trans = self.__cal_jac_trans(self.__trans_end_in_last)

        ### gravity vector
        self.__model.gravity.linear = gravity

    def get_gravity(self) -> np.ndarray:
        return copy.deepcopy(self.__model.gravity.linear)

    def set_gravity(
            self,
            gravity: np.ndarray = np.array([0, 0, -9.81]),
    ):
        self.__model.gravity.linear = copy.deepcopy(gravity)

    def get_joint_num(self) -> int:
        return self.__joint_num

    def __cal_jac_trans(self, trans_end_in_last: np.ndarray) -> np.ndarray:
        rot = trans_end_in_last[:3, :3]
        pos = trans_end_in_last[:3, 3]

        jac_trans = np.eye(6)
        jac_trans[:3, :3] = rot.T
        jac_trans[3:, :3] = -rot.T @ hat(pos)
        jac_trans[3:, 3:] = rot.T
        return jac_trans

    # get [M(q), C(q, q_dot), G(q), J(q), J_dot(q, q_dot)]
    # v = J @ q_dot
    def dynamic_params(
        self,
        q: np.ndarray,
        dq: np.ndarray,
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        # Compute all dynamic parameters
        pin.computeAllTerms(self.__model, self.__data, q, dq)
        m_mat = self.__data.M
        c_mat = pin.computeCoriolisMatrix(self.__model, self.__data, q, dq)
        g_vec = self.__data.g
        jac = pin.computeJointJacobian(
            self.__model,
            self.__data,
            q,
            self.__end_joint_id,
        )
        jac_dot = pin.computeJointJacobiansTimeVariation(
            self.__model,
            self.__data,
            q,
            dq,
        )

        jac = self.__jac_trans @ jac
        jac_dot = self.__jac_trans @ jac_dot
        return m_mat, c_mat, g_vec, jac, jac_dot

    # get [pose_1, pose_2, ..., pose_n]
    def forward_kinematics(
        self,
        q: np.ndarray,
    ) -> List[Tuple[np.ndarray, np.ndarray]]:

        # Compute forward kinematics to update joint placements
        pin.forwardKinematics(self.__model, self.__data, q)

        # Collect the poses of all joints
        poses = []
        trans = None
        for i in range(self.__joint_num):
            trans = self.__data.oMi[i + 1].homogeneous
            pos, quat = trans2part(trans)
            poses.append((pos, quat))
        trans_end_in_base = trans @ self.__trans_end_in_last
        pos, quat = trans2part(trans_end_in_base)
        poses.append((pos, quat))
        return poses

    def inverse_kinematics(
        self,
        tar_pose: Tuple[np.ndarray, np.ndarray],
        start_q: np.ndarray,
        dt: float = 1e-1,
        exit_eps: float = 1e-3,
        feasible_eps: float = 1e-2,
        damp: float = 1e-12,
        max_iter: int = 300,
    ) -> Tuple[bool, np.ndarray, float]:
        result_q = copy.deepcopy(start_q)
        trans_end_tar_in_base = copy.deepcopy(
            part2trans(
                tar_pose[0],
                tar_pose[1],
            ))
        trans_tar_in_base = trans_end_tar_in_base @ self.__trans_last_in_end
        trans_base_in_tar = trans_inv(trans_tar_in_base)

        # inverse kinematics
        result_flag = False
        for _ in range(max_iter):
            pin.forwardKinematics(self.__model, self.__data, result_q)
            trans_end_in_base = self.__data.oMi[
                self.__end_joint_id].homogeneous
            trans_tar_in_end = trans_inv(trans_end_in_base) @ trans_tar_in_base
            err = trans2se3(trans_tar_in_end)

            err_norm = np.linalg.norm(err)
            if err_norm < exit_eps:
                result_flag = True
                break

            # jac in joint frame
            jac = pin.computeJointJacobian(
                self.__model,
                self.__data,
                result_q,
                self.__end_joint_id,
            )
            dq = np.linalg.pinv(jac, rcond=damp) @ err
            result_q = pin.integrate(
                self.__model,
                result_q,
                dq * dt,
            )

        # post process
        result_q = np.clip(
            angle_norm(result_q),
            self.__lower_limit,
            self.__upper_limit,
        )

        # check feasible
        pin.forwardKinematics(self.__model, self.__data, result_q)
        trans_end_in_base = self.__data.oMi[self.__end_joint_id].homogeneous
        trans_tar_in_end = trans_base_in_tar @ trans_end_in_base
        err = trans2se3(trans_tar_in_end)
        err_norm = np.linalg.norm(err)
        if err_norm < feasible_eps:
            result_flag = True

        return result_flag, result_q, err_norm


class HexFricUtil:
    # tanh-based friction model
    # tau_f = fc * tanh(k * dq) + fv * dq + fo
    def __init__(
            self,
            fc: np.ndarray = np.array([1.0] * 6),
            k: np.ndarray = np.array([100.0] * 6),
            fv: np.ndarray = np.array([1.0] * 6),
            fo: np.ndarray = np.array([0.0] * 6),
    ):
        # constants
        self.__fc = fc.copy()
        self.__k = k.copy()
        self.__fv = fv.copy()
        self.__fo = fo.copy()

    def __call__(self, dq: np.ndarray):
        tau_c = self.__fc * np.tanh(self.__k * dq)
        tau_v = self.__fv * dq
        tau_o = self.__fo
        tau_f = tau_c + tau_v + tau_o
        return tau_f



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
################################################################
# Copyright 2025 Dong Zhaorui. All rights reserved.
# Author: Dong Zhaorui 847235539@qq.com
# Date  : 2025-09-19
################################################################

import time
import h5py
import numpy as np


class HexHdf5Reader:

    def __init__(self, file_path: str):
        self.__file_path = file_path
        self.__hdf5_file = h5py.File(file_path, "r")
        self.__dataset_dict = {}

    def __del__(self):
        if hasattr(self, '__hdf5_file') and self.__hdf5_file is not None:
            self.__hdf5_file.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        if self.__hdf5_file is not None:
            self.__hdf5_file.close()
            self.__hdf5_file = None

    def get_shape(self, group_name: str, dataset_name: str = "data"):
        dataset = self.__get_dataset_handle(group_name, dataset_name)
        return dataset.shape

    def get_dtype(self, group_name: str, dataset_name: str = "data"):
        dataset = self.__get_dataset_handle(group_name, dataset_name)
        return dataset.dtype

    def summary(self):
        if self.__hdf5_file is None:
            print("HDF5 file is closed. Cannot get summary.")
            return

        print("#" * 50)
        print(f"HDF5 File: {self.__file_path}")
        print("#" * 50)

        for group_name in self.__hdf5_file.keys():
            print("-" * 30)
            print(f"Group: {group_name}")
            print("-" * 30)
            print(
                f"get_delta: {(self.__hdf5_file[group_name]['get_ts'][-1] - self.__hdf5_file[group_name]['get_ts'][0]) / 1e9}s"
            )
            print(
                f"sen_delta: {(self.__hdf5_file[group_name]['sen_ts'][-1] - self.__hdf5_file[group_name]['sen_ts'][0]) / 1e9}s"
            )

            dtype = self.get_dtype(group_name)
            shape = self.get_shape(group_name)
            print(f"  Dataset: {group_name}")
            print(f"    Dtype: {dtype}")
            print(f"    Shape: {shape}")

        print("#" * 50)

    def __get_dataset_handle(
        self,
        group_name: str,
        dataset_name: str = "data",
    ):
        dataset_key = f"{group_name}/{dataset_name}"
        if dataset_key not in self.__dataset_dict:
            if group_name not in self.__hdf5_file:
                raise KeyError(f"Group '{group_name}' not found in HDF5 file")
            if dataset_name not in self.__hdf5_file[group_name]:
                raise KeyError(
                    f"Dataset '{dataset_name}' not found in group '{group_name}'"
                )
            self.__dataset_dict[dataset_key] = self.__hdf5_file[group_name][
                dataset_name]
        return self.__dataset_dict[dataset_key]

    def get_data(
        self,
        group_name: str,
        index: int,
        dataset_name: str = "data",
        use_ns: bool = False,
    ):
        return self.get_batch_data(
            group_name=group_name,
            start_index=index,
            end_index=index + 1,
            dataset_name=dataset_name,
            use_ns=use_ns,
        )[0]

    def get_all_data(
        self,
        group_name: str,
        dataset_name: str = "data",
        use_ns: bool = False,
    ):
        return self.get_batch_data(
            group_name=group_name,
            start_index=0,
            end_index=None,
            dataset_name=dataset_name,
            use_ns=use_ns,
        )

    def get_batch_data(
        self,
        group_name: str,
        start_index: int,
        end_index: int = None,
        dataset_name: str = "data",
        use_ns: bool = False,
    ):
        dataset = self.__get_dataset_handle(group_name, dataset_name)

        if start_index < 0:
            start_index = 0
        elif start_index >= dataset.shape[0]:
            raise IndexError(
                f"Start index {start_index} out of range [0, {dataset.shape[0]})"
            )

        if end_index is None or end_index > dataset.shape[0]:
            end_index = dataset.shape[0]
        elif end_index < start_index:
            end_index = start_index

        data = dataset[start_index:end_index]
        if dataset_name != "data":
            if data.shape[1:] != (1, ):
                raise ValueError(
                    f"Dataset {group_name}/{dataset_name} is not a timestamp dataset (expected shape[1:] = (1,), got {data.shape[1:]})"
                )
            if not use_ns:
                ts_list = []
                for ts_ns in data.reshape(-1):
                    ts_list.append(self.__ns_to_hex_ts(int(ts_ns)))
                return ts_list
            else:
                return data.reshape(-1)
        return data

    def __ns_to_hex_ts(self, ts: int):
        return {
            "s": ts // 1_000_000_000,
            "ns": ts % 1_000_000_000,
        }

    def now_ns(self):
        return np.array([time.time_ns()])

    def hex_ts_to_ns(self, ts: dict):
        try:
            return np.array([ts["s"] * 1e9 + ts["ns"]])
        except Exception as e:
            print(f"hex_ts_to_ns failed: {e}")
            return np.array([np.inf])
